@@ -27,10 +27,7 @@
 #include <rccl.h>
 #endif
 
-#if HAVE_MPI
-#include "../mpi/mpi_context.h"
-#endif
-
+#include "../hashes.h"
 #include "gpu_operations.h"
 
 #include <functional>
@@ -41,21 +38,26 @@ namespace common {
 ncclDataType_t GetNCCLDataType(const std::shared_ptr<Tensor> tensor);
 
 struct NCCLContext {
-  std::vector<std::unordered_map<std::vector<int32_t>, ncclComm_t>> nccl_comms;
+  // indexed by [nccl stream][{process set id, device id vector}]
+  std::vector<
+      std::unordered_map<std::tuple<int32_t, std::vector<int32_t>>, ncclComm_t>>
+      nccl_comms;
 
-  void ErrorCheck(std::string op_name, ncclResult_t nccl_result, ncclComm_t& nccl_comm);
+  void ErrorCheck(std::string op_name, ncclResult_t nccl_result,
+                  ncclComm_t& nccl_comm);
 
   void ShutDown();
+
+  bool elastic;
 };
 
 class NCCLOpContext {
 public:
   NCCLOpContext(NCCLContext* nccl_context, HorovodGlobalState* global_state,
-                horovod::common::Communicator communicator_type)
+                Communicator communicator_type)
       : nccl_comm_(nullptr),
         error_check_callback_(std::bind(&NCCLOpContext::AsyncErrorCheck, this)),
-        nccl_context_(nccl_context),
-        global_state_(global_state),
+        nccl_context_(nccl_context), global_state_(global_state),
         communicator_type_(communicator_type){};
 
   void InitNCCLComm(const std::vector<TensorTableEntry>& entries,
@@ -68,20 +70,20 @@ public:
 
 private:
   void PopulateNCCLCommStrategy(int& nccl_rank, int& nccl_size,
-                                Communicator& nccl_id_bcast_comm);
+                                Communicator& nccl_id_bcast_comm,
+                                const ProcessSet& process_set);
 
   NCCLContext* nccl_context_;
   HorovodGlobalState* global_state_;
-  horovod::common::Communicator communicator_type_;
+  Communicator communicator_type_;
 };
 
 class NCCLAllreduce : public GPUAllreduce {
 public:
   NCCLAllreduce(NCCLContext* nccl_context, GPUContext* gpu_context,
                 HorovodGlobalState* global_state,
-                horovod::common::Communicator communicator_type = Communicator::GLOBAL)
-      : GPUAllreduce(gpu_context, global_state),
-        nccl_context_(nccl_context),
+                Communicator communicator_type = Communicator::GLOBAL)
+      : GPUAllreduce(gpu_context, global_state), nccl_context_(nccl_context),
         nccl_op_context_(nccl_context, global_state, communicator_type),
         global_state_(global_state){};
 
@@ -89,6 +91,8 @@ public:
                  const Response& response) override;
 
 protected:
+  void WaitForData(std::vector<TensorTableEntry>& entries) override;
+
   NCCLContext* nccl_context_;
   NCCLOpContext nccl_op_context_;
   HorovodGlobalState* global_state_;
@@ -98,8 +102,7 @@ class NCCLBroadcast : public GPUBroadcast {
 public:
   NCCLBroadcast(NCCLContext* nccl_context, GPUContext* gpu_context,
                 HorovodGlobalState* global_state)
-      : GPUBroadcast(gpu_context, global_state),
-        nccl_context_(nccl_context),
+      : GPUBroadcast(gpu_context, global_state), nccl_context_(nccl_context),
         nccl_op_context_(nccl_context, global_state, Communicator::GLOBAL),
         global_state_(global_state){};
 
@@ -107,6 +110,8 @@ public:
                  const Response& response) override;
 
 protected:
+  void WaitForData(std::vector<TensorTableEntry>& entries) override;
+
   NCCLContext* nccl_context_;
   NCCLOpContext nccl_op_context_;
   HorovodGlobalState* global_state_;
@@ -116,8 +121,7 @@ class NCCLAlltoall : public GPUAlltoall {
 public:
   NCCLAlltoall(NCCLContext* nccl_context, GPUContext* gpu_context,
                HorovodGlobalState* global_state)
-      : GPUAlltoall(gpu_context, global_state),
-        nccl_context_(nccl_context),
+      : GPUAlltoall(gpu_context, global_state), nccl_context_(nccl_context),
         nccl_op_context_(nccl_context, global_state, Communicator::GLOBAL),
         global_state_(global_state){};
 
@@ -125,6 +129,8 @@ public:
                  const Response& response) override;
 
 protected:
+  void WaitForData(std::vector<TensorTableEntry>& entries) override;
+
   NCCLContext* nccl_context_;
   NCCLOpContext nccl_op_context_;
   HorovodGlobalState* global_state_;
@@ -133,11 +139,10 @@ protected:
 #if HAVE_MPI
 class NCCLHierarchicalAllreduce : public NCCLAllreduce {
 public:
-  NCCLHierarchicalAllreduce(NCCLContext* nccl_context, MPIContext* mpi_context,
-                            GPUContext* gpu_context,
+  NCCLHierarchicalAllreduce(NCCLContext* nccl_context, GPUContext* gpu_context,
                             HorovodGlobalState* global_state)
-      : NCCLAllreduce(nccl_context, gpu_context, global_state, Communicator::LOCAL),
-        mpi_context_(mpi_context){};
+      : NCCLAllreduce(nccl_context, gpu_context, global_state,
+                      Communicator::LOCAL){};
 
   Status Execute(std::vector<TensorTableEntry>& entries,
                  const Response& response) override;
@@ -146,6 +151,9 @@ public:
                const std::vector<TensorTableEntry>& entries,
                const Response& response) const override;
 
+protected:
+  void WaitForData(std::vector<TensorTableEntry>& entries) override;
+
 private:
   MPIContext* mpi_context_;
 };
@@ -154,10 +162,10 @@ private:
 class NCCLAllgather : public GPUAllgather {
 public:
   NCCLAllgather(NCCLContext* nccl_context, GPUContext* gpu_context,
-                  HorovodGlobalState* global_state)
-      : GPUAllgather(gpu_context, global_state),
+                HorovodGlobalState* global_state)
+      : GPUAllgather(gpu_context, global_state), nccl_context_(nccl_context),
         nccl_op_context_(nccl_context, global_state, Communicator::GLOBAL),
-        global_state_(global_state){};
+        global_state_(global_state) {}
 
   Status Execute(std::vector<TensorTableEntry>& entries,
                  const Response& response) override;
@@ -167,12 +175,13 @@ public:
                const Response& response) const override;
 
 protected:
+  void WaitForData(std::vector<TensorTableEntry>& entries) override;
+
   NCCLContext* nccl_context_;
   NCCLOpContext nccl_op_context_;
   HorovodGlobalState* global_state_;
 };
 
-
 } // namespace common
 } // namespace horovod
 
