@@ -27,6 +27,7 @@ OperationManager::OperationManager(ParameterManager* param_manager,
                                    std::vector<std::shared_ptr<AlltoallOp>> alltoall_ops,
                                    std::shared_ptr<JoinOp> join_op,
                                    std::vector<std::shared_ptr<AllreduceOp>> adasum_ops,
+                                   std::shared_ptr<BarrierOp> barrier_op,
                                    std::shared_ptr<ErrorOp> error_op)
     : param_manager_(param_manager),
       allreduce_ops_(std::move(allreduce_ops)),
@@ -35,6 +36,7 @@ OperationManager::OperationManager(ParameterManager* param_manager,
       alltoall_ops_(std::move(alltoall_ops)),
       join_op_(std::move(join_op)),
       adasum_ops_(std::move(adasum_ops)),
+      barrier_op_(std::move(barrier_op)),
       error_op_(std::move(error_op)) {}
 
 Status OperationManager::ExecuteAllreduce(std::vector<TensorTableEntry>& entries,
@@ -78,8 +80,9 @@ Status OperationManager::ExecuteAlltoall(std::vector<TensorTableEntry>& entries,
 }
 
 Status OperationManager::ExecuteJoin(std::vector<TensorTableEntry>& entries,
-                                          const Response& response) const {
-  return join_op_->Execute(entries, response);
+                                     const Response& response,
+                                     ProcessSet& process_set) const {
+  return join_op_->Execute(entries, response, process_set);
 }
 
 Status OperationManager::ExecuteAdasum(std::vector<TensorTableEntry>& entries,
@@ -92,13 +95,19 @@ Status OperationManager::ExecuteAdasum(std::vector<TensorTableEntry>& entries,
   throw std::logic_error("No Adasum operation enabled");
 }
 
+Status OperationManager::ExecuteBarrier(std::vector<TensorTableEntry>& entries,
+                                     const Response& response) const {
+  return barrier_op_->Execute(entries, response);
+}
+
 Status OperationManager::ExecuteError(std::vector<TensorTableEntry>& entries,
                                       const Response& response) const {
   return error_op_->Execute(entries, response);
 }
 
 Status OperationManager::ExecuteOperation(std::vector<TensorTableEntry>& entries,
-                                          const Response& response) const {
+                                          const Response& response,
+                                          ProcessSet& process_set) const {
   if (response.response_type() == Response::ALLREDUCE) {
     return ExecuteAllreduce(entries, response);
   } else if (response.response_type() == Response::ALLGATHER) {
@@ -108,9 +117,11 @@ Status OperationManager::ExecuteOperation(std::vector<TensorTableEntry>& entries
   } else if (response.response_type() == Response::ALLTOALL) {
     return ExecuteAlltoall(entries, response);
   } else if (response.response_type() == Response::JOIN) {
-    return ExecuteJoin(entries, response);
+    return ExecuteJoin(entries, response, process_set);
   } else if (response.response_type() == Response::ADASUM) {
     return ExecuteAdasum(entries, response);
+  } else if (response.response_type() == Response::BARRIER) {
+    return ExecuteBarrier(entries, response);
   } else if (response.response_type() == Response::ERROR) {
     return ExecuteError(entries, response);
   } else {
