@@ -16,11 +16,11 @@
 import sys
 
 from horovod.spark.task import task_exec
-from horovod.run.common.util import codec
+from horovod.runner.common.util import codec
 
 
 def main(driver_addresses, settings):
-    task_exec(driver_addresses, settings, 'HOROVOD_RANK')
+    task_exec(driver_addresses, settings, 'HOROVOD_RANK', 'HOROVOD_LOCAL_RANK')
 
 
 if __name__ == '__main__':
