@@ -18,14 +18,27 @@
 
 #include <memory>
 #include <queue>
+#include <regex>
 #include <thread>
 #include <unordered_map>
 
+#define EIGEN_USE_THREADS
+#if HAVE_CUDA || HAVE_ROCM
+#define EIGEN_USE_GPU
+#endif  // HAVE_CUDA || HAVE_ROCM
+
 #include "tensorflow/core/framework/op.h"
 #include "tensorflow/core/framework/op_kernel.h"
 #include "tensorflow/core/framework/shape_inference.h"
+#include "tensorflow/core/framework/common_shape_fns.h"
 
-#define EIGEN_USE_THREADS
+#if TENSORFLOW_VERSION >= 2006000000
+#include "tensorflow/core/framework/resource_mgr.h"
+#include "tensorflow/core/framework/resource_var.h"
+#include "tensorflow/core/kernels/training_op_helpers.h"
+#endif // TENSORFLOW_VERSION >= 2006000000
+
+#include "../common/common.h"
 
 #if HAVE_GPU
 
@@ -120,14 +133,26 @@ common::Status ConvertStatus(const Status& status) {
   }
 }
 
+int GetDeviceID(OpKernelContext* context);
+
 #if HAVE_GPU
+struct ReadyEventRegistry {
+  std::unordered_map<int, std::queue<gpuEvent_t>> gpu_events;
+  std::mutex mutex;
+};
+
+static ReadyEventRegistry ready_event_registry;
+
 class TFReadyEvent : public common::ReadyEvent {
 public:
-  TFReadyEvent(DeviceContext* device_context);
+  TFReadyEvent(OpKernelContext* context);
+  ~TFReadyEvent();
   bool Ready() const override;
+  gpuEvent_t event() const override;
 
 private:
-  std::shared_ptr<perftools::gputools::Event> event_;
+  gpuEvent_t event_;
+  int device_ = CPU_DEVICE_ID;
 };
 #endif
 
@@ -138,7 +163,7 @@ public:
   AccessData(std::shared_ptr<common::OpContext> context) const override;
 
 private:
-  std::shared_ptr<PersistentTensor> tensor_;
+  std::shared_ptr<Tensor> tensor_;
 };
 
 class TFTensor : public common::Tensor {
@@ -176,27 +201,47 @@ private:
 };
 
 #if HAVE_GPU
-TFReadyEvent::TFReadyEvent(DeviceContext* device_context) {
-  auto executor = device_context->stream()->parent();
-  auto ready_event = new perftools::gputools::Event(executor);
-  ready_event->Init();
-  device_context->stream()->ThenRecordEvent(ready_event);
-  event_ = std::shared_ptr<perftools::gputools::Event>(ready_event);
+TFReadyEvent::TFReadyEvent(OpKernelContext* context) {
+  device_ = GetDeviceID(context);
+  {
+    std::lock_guard<std::mutex> guard(ready_event_registry.mutex);
+    auto& queue = ready_event_registry.gpu_events[device_];
+    if (!queue.empty()) {
+      event_ = queue.front();
+      queue.pop();
+    } else {
+      HVD_GPU_CHECK(gpuEventCreateWithFlags(&event_, gpuEventDisableTiming));
+    }
+  }
+  auto device_context = context->op_device_context();
+  auto stream = stream_executor::gpu::AsGpuStreamValue(device_context->stream());
+  HVD_GPU_CHECK(gpuEventRecord(event_, stream));
 }
 
 bool TFReadyEvent::Ready() const {
-  return event_->PollForStatus() !=
-         perftools::gputools::Event::Status::kPending;
+  HVD_GPU_CHECK(gpuEventSynchronize(event_));
+  return true;
 }
+
+gpuEvent_t TFReadyEvent::event() const {
+  return event_;
+}
+
+TFReadyEvent::~TFReadyEvent() {
+  {
+    std::lock_guard<std::mutex> guard(ready_event_registry.mutex);
+    auto& queue = ready_event_registry.gpu_events[device_];
+    queue.push(event_);
+  }
+}
+
 #endif
 
 TFPersistentBuffer::TFPersistentBuffer(OpKernelContext* context, int64_t size) {
-  tensor_ = std::make_shared<PersistentTensor>();
+  tensor_ = std::make_shared<Tensor>();
   TensorShape buffer_shape;
   buffer_shape.AddDim(size);
-  Tensor* unused;
-  Status status = context->allocate_persistent(DT_INT8, buffer_shape,
-                                               tensor_.get(), &unused);
+  Status status = context->allocate_temp(DT_INT8, buffer_shape, tensor_.get());
   if (!status.ok()) {
     throw status;
   }
@@ -212,13 +257,7 @@ TFPersistentBuffer::TFPersistentBuffer(OpKernelContext* context, int64_t size) {
 
 const void* TFPersistentBuffer::AccessData(
     std::shared_ptr<common::OpContext> context) const {
-  // It's safe to cast context to TFOpContext, since only TFOpContext creates
-  // TFPersistentBuffer.
-  return (const void *)tensor_
-      ->AccessTensor(
-          std::dynamic_pointer_cast<TFOpContext>(context)->GetKernelContext())
-      ->tensor_data()
-      .data();
+  return (const void *)tensor_->tensor_data().data();
 }
 
 TFTensor::TFTensor(::tensorflow::Tensor& tensor) : tensor_(tensor) {}
@@ -310,8 +349,7 @@ int GetDeviceID(OpKernelContext* context);
 common::Status
 TFOpContext::AllocateZeros(int64_t num_elements, common::DataType dtype,
                            std::shared_ptr<common::Tensor>* tensor) {
-  ::tensorflow::Tensor* unused;
-  std::shared_ptr<PersistentTensor> zero_tensor = std::make_shared<PersistentTensor>();
+  std::shared_ptr<Tensor> zero_tensor = std::make_shared<Tensor>();
   auto tf_data_type = GetTFDataType(dtype);
   ::tensorflow::AllocatorAttributes tf_attribute;
   int device_ = GetDeviceID(context_);
@@ -322,22 +360,21 @@ TFOpContext::AllocateZeros(int64_t num_elements, common::DataType dtype,
     tf_attribute.set_on_host(true);
   }
 
-  Status status = context_->allocate_persistent(tf_data_type, ::tensorflow::TensorShape({num_elements}), zero_tensor.get(), &unused, tf_attribute);
+  Status status = context_->allocate_temp(tf_data_type, ::tensorflow::TensorShape({num_elements}), zero_tensor.get(), tf_attribute);
 
   if (device_ != CPU_DEVICE_ID) {
 #if HAVE_GPU
     auto device_context = context_->op_device_context();
     auto stream = (device_context != nullptr) ? stream_executor::gpu::AsGpuStreamValue(device_context->stream()) : 0;
-    void *ptr = (void*)zero_tensor->AccessTensor(hvd_context->GetKernelContext())->tensor_data().data();
-    auto size = zero_tensor->AccessTensor(hvd_context->GetKernelContext())->tensor_data().size();
+    void *ptr = (void*)zero_tensor->tensor_data().data();
+    auto size = zero_tensor->tensor_data().size();
     gpuMemsetAsync(ptr, 0, size, stream);
 #endif
   } else {
-    memset((void*)zero_tensor->AccessTensor(hvd_context->GetKernelContext())->tensor_data().data(), 0,
-           zero_tensor->AccessTensor(hvd_context->GetKernelContext())->tensor_data().size());
+    memset((void*)zero_tensor->tensor_data().data(), 0, zero_tensor->tensor_data().size());
   }
   if (status.ok()) {
-    *tensor = std::make_shared<TFTensor>(*(zero_tensor->AccessTensor(hvd_context->GetKernelContext())));
+    *tensor = std::make_shared<TFTensor>(*zero_tensor);
   }
 
 #if HAVE_GPU
@@ -368,15 +405,15 @@ int GetDeviceID(OpKernelContext* context) {
 
 // On GPU this event will signal that data is ready, and tensors are
 // allocated.
-common::ReadyEvent* RecordReadyEvent(OpKernelContext* context) {
 #if HAVE_GPU
+common::ReadyEvent* RecordReadyEvent(OpKernelContext* context) {
   auto device_context = context->op_device_context();
   if (device_context != nullptr) {
-    return new TFReadyEvent(device_context);
+    return new TFReadyEvent(context);
   }
-#endif
   return nullptr;
 }
+#endif
 
 } // namespace
 
@@ -388,6 +425,7 @@ public:
     OP_REQUIRES_OK(context, context->GetAttr("prescale_factor", &prescale_factor_));
     OP_REQUIRES_OK(context, context->GetAttr("postscale_factor", &postscale_factor_));
     OP_REQUIRES_OK(context, context->GetAttr("ignore_name_scope", &ignore_name_scope_));
+    OP_REQUIRES_OK(context, context->GetAttr("process_set_id", &process_set_id_));
   }
 
   void ComputeAsync(OpKernelContext* context, DoneCallback done) override {
@@ -408,16 +446,31 @@ public:
     OP_REQUIRES_OK_ASYNC(
         context, context->allocate_output(0, tensor.shape(), &output), done);
     // ReadyEvent makes sure input tensor is ready, and output is allocated.
-    auto ready_event = std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context));
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
     auto hvd_context = std::make_shared<TFOpContext>(context);
     auto hvd_tensor = std::make_shared<TFTensor>(tensor);
     auto hvd_output = std::make_shared<TFTensor>(*output);
     auto enqueue_result = EnqueueTensorAllreduce(
-        hvd_context, hvd_tensor, hvd_output, ready_event, node_name, device,
+        hvd_context, hvd_tensor, hvd_output, ready_event_list, node_name, device,
         [context, done](const common::Status& status) {
+#if HAVE_GPU
+          auto hvd_event = status.event;
+          if (hvd_event.event) {
+            auto device_context = context->op_device_context();
+            if (device_context != nullptr) {
+                auto stream = stream_executor::gpu::AsGpuStreamValue(device_context->stream());
+                HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+            }
+          }
+#endif
           context->SetStatus(ConvertStatus(status));
           done();
-        }, reduce_op, (double) prescale_factor_, (double) postscale_factor_);
+        },
+        reduce_op, (double)prescale_factor_, (double)postscale_factor_,
+        process_set_id_);
     OP_REQUIRES_OK_ASYNC(context, ConvertStatus(enqueue_result), done);
   }
 
@@ -427,6 +480,7 @@ private:
   float prescale_factor_;
   float postscale_factor_;
   bool ignore_name_scope_;
+  int process_set_id_;
 };
 
 REGISTER_KERNEL_BUILDER(Name("HorovodAllreduce").Device(DEVICE_CPU),
@@ -442,6 +496,7 @@ REGISTER_OP("HorovodAllreduce")
     .Attr("prescale_factor: float")
     .Attr("postscale_factor: float")
     .Attr("ignore_name_scope: bool = False")
+    .Attr("process_set_id: int = 0")
     .Input("tensor: T")
     .Output("sum: T")
     .SetShapeFn([](shape_inference::InferenceContext* c) {
@@ -470,6 +525,7 @@ public:
     OP_REQUIRES_OK(context, context->GetAttr("postscale_factor", &postscale_factor_));
     OP_REQUIRES_OK(context, context->GetAttr("ignore_name_scope", &ignore_name_scope_));
     OP_REQUIRES_OK(context, context->GetAttr("num_tensors", &num_tensors_));
+    OP_REQUIRES_OK(context, context->GetAttr("process_set_id", &process_set_id_));
   }
 
   void ComputeAsync(OpKernelContext* context, DoneCallback done) override {
@@ -487,13 +543,13 @@ public:
     horovod::common::ReduceOp reduce_op = static_cast<horovod::common::ReduceOp>(reduce_op_);
     std::vector<Tensor*> outputs(num_tensors_);
 
-    std::vector<std::shared_ptr<common::ReadyEvent>> ready_events;
+    std::vector<common::ReadyEventList> ready_event_lists;
     std::vector<std::shared_ptr<common::OpContext>> hvd_contexts;
     std::vector<std::shared_ptr<common::Tensor>> hvd_tensors;
     std::vector<std::shared_ptr<common::Tensor>> hvd_outputs;
     std::vector<common::StatusCallback> callbacks;
     std::vector<std::string> names;
-    ready_events.reserve(num_tensors_);
+    ready_event_lists.reserve(num_tensors_);
     hvd_contexts.reserve(num_tensors_);
     hvd_tensors.reserve(num_tensors_);
     hvd_outputs.reserve(num_tensors_);
@@ -508,9 +564,17 @@ public:
       OP_REQUIRES_OK_ASYNC(
           context, context->allocate_output(i, tensor.shape(), &outputs[i]),
           done);
-      // ReadyEvent makes sure input tensor is ready, and output is allocated.
-      ready_events.emplace_back(
-          std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+    }
+
+    // ReadyEvent makes sure input tensors are ready, and outputs are allocated.
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
+
+    for (int i = 0; i < num_tensors_; ++i) {
+      auto tensor = context->input(i);
+      ready_event_lists.emplace_back(ready_event_list); // Same for all tensors in group
       hvd_contexts.emplace_back(std::make_shared<TFOpContext>(context));
       hvd_tensors.emplace_back(std::make_shared<TFTensor>(tensor));
       names.emplace_back(node_name + "_" + std::to_string(i + 1) + "of" +
@@ -523,6 +587,16 @@ public:
             std::lock_guard<std::mutex> guard(*callback_mutex);
             (*callback_count)++;
             if (*callback_count == num_tensors) {
+#if HAVE_GPU
+              auto hvd_event = status.event;
+              if (hvd_event.event) {
+                auto device_context = context->op_device_context();
+                if (device_context != nullptr) {
+                    auto stream = stream_executor::gpu::AsGpuStreamValue(device_context->stream());
+                    HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+                }
+              }
+#endif
               context->SetStatus(ConvertStatus(status));
               done();
             }
@@ -530,8 +604,9 @@ public:
     }
 
     auto enqueue_result = EnqueueTensorAllreduces(
-        hvd_contexts, hvd_tensors, hvd_outputs, ready_events, names, device,
-        callbacks, reduce_op, (double) prescale_factor_, (double) postscale_factor_);
+        hvd_contexts, hvd_tensors, hvd_outputs, ready_event_lists, names, device,
+        callbacks, reduce_op, (double)prescale_factor_,
+        (double)postscale_factor_, process_set_id_);
     OP_REQUIRES_OK_ASYNC(context, ConvertStatus(enqueue_result), done);
   }
 
@@ -542,6 +617,7 @@ private:
   float postscale_factor_;
   bool ignore_name_scope_;
   int num_tensors_;
+  int process_set_id_;
 };
 
 REGISTER_KERNEL_BUILDER(Name("HorovodGroupedAllreduce").Device(DEVICE_CPU),
@@ -558,6 +634,7 @@ REGISTER_OP("HorovodGroupedAllreduce")
     .Attr("postscale_factor: float")
     .Attr("ignore_name_scope: bool = False")
     .Attr("num_tensors: int")
+    .Attr("process_set_id: int = 0")
     .Input("tensors: num_tensors*T")
     .Output("sum: num_tensors*T")
     .SetShapeFn([](shape_inference::InferenceContext* c) {
@@ -584,6 +661,7 @@ public:
   explicit HorovodAllgatherOp(OpKernelConstruction* context)
       : AsyncOpKernel(context) {
     OP_REQUIRES_OK(context, context->GetAttr("ignore_name_scope", &ignore_name_scope_));
+    OP_REQUIRES_OK(context, context->GetAttr("process_set_id", &process_set_id_));
   }
 
   void ComputeAsync(OpKernelContext* context, DoneCallback done) override {
@@ -602,20 +680,35 @@ public:
     // ReadyEvent makes sure input tensor is ready.  We cannot pre-allocate
     // output for allgather, since shape of result is only known after all
     // ranks make a request.
-    auto ready_event = std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context));
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
     auto hvd_context = std::make_shared<TFOpContext>(context);
     auto hvd_tensor = std::make_shared<TFTensor>(tensor);
     auto enqueue_result = EnqueueTensorAllgather(
-        hvd_context, hvd_tensor, ready_event, node_name, device,
+        hvd_context, hvd_tensor, ready_event_list, node_name, device,
         [context, done](const common::Status& status) {
+#if HAVE_GPU
+          auto hvd_event = status.event;
+          if (hvd_event.event) {
+            auto device_context = context->op_device_context();
+            if (device_context != nullptr) {
+                auto stream = stream_executor::gpu::AsGpuStreamValue(device_context->stream());
+                HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+            }
+          }
+#endif
           context->SetStatus(ConvertStatus(status));
           done();
-        });
+        },
+        process_set_id_);
     OP_REQUIRES_OK_ASYNC(context, ConvertStatus(enqueue_result), done);
   }
 
 private:
   bool ignore_name_scope_;
+  int process_set_id_;
 };
 
 REGISTER_KERNEL_BUILDER(Name("HorovodAllgather").Device(DEVICE_CPU),
@@ -629,6 +722,7 @@ REGISTER_OP("HorovodAllgather")
     .Attr(
         "T: {uint8, int8, uint16, int16, int32, int64, float16, float32, float64, bool}")
     .Attr("ignore_name_scope: bool = False")
+    .Attr("process_set_id: int = 0")
     .Input("tensor: T")
     .Output("output: T")
     .SetShapeFn([](shape_inference::InferenceContext* c) {
@@ -656,6 +750,7 @@ public:
       : AsyncOpKernel(context) {
     OP_REQUIRES_OK(context, context->GetAttr("root_rank", &root_rank_));
     OP_REQUIRES_OK(context, context->GetAttr("ignore_name_scope", &ignore_name_scope_));
+    OP_REQUIRES_OK(context, context->GetAttr("process_set_id", &process_set_id_));
   }
 
   void ComputeAsync(OpKernelContext* context, DoneCallback done) override {
@@ -679,7 +774,10 @@ public:
           context, context->allocate_output(0, tensor.shape(), &output), done);
     }
     // ReadyEvent makes sure input tensor is ready, and output is allocated.
-    auto ready_event = std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context));
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
     auto hvd_context = std::make_shared<TFOpContext>(context);
     auto hvd_tensor = std::make_shared<TFTensor>(tensor);
     std::shared_ptr<TFTensor> hvd_output = nullptr;
@@ -687,17 +785,29 @@ public:
       hvd_output = std::make_shared<TFTensor>(*output);
     }
     auto enqueue_result = EnqueueTensorBroadcast(
-        hvd_context, hvd_tensor, hvd_output, root_rank_, ready_event, node_name,
+        hvd_context, hvd_tensor, hvd_output, root_rank_, ready_event_list, node_name,
         device, [context, done](const common::Status& status) {
+#if HAVE_GPU
+          auto hvd_event = status.event;
+          if (hvd_event.event) {
+            auto device_context = context->op_device_context();
+            if (device_context != nullptr) {
+                auto stream = stream_executor::gpu::AsGpuStreamValue(device_context->stream());
+                HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+            }
+          }
+#endif
           context->SetStatus(ConvertStatus(status));
           done();
-        });
+        },
+        process_set_id_);
     OP_REQUIRES_OK_ASYNC(context, ConvertStatus(enqueue_result), done);
   }
 
 private:
   int root_rank_;
   bool ignore_name_scope_;
+  int process_set_id_;
 };
 
 REGISTER_KERNEL_BUILDER(Name("HorovodBroadcast").Device(DEVICE_CPU),
@@ -712,6 +822,7 @@ REGISTER_OP("HorovodBroadcast")
         "T: {uint8, int8, uint16, int16, int32, int64, float16, float32, float64, bool}")
     .Attr("root_rank: int")
     .Attr("ignore_name_scope: bool = False")
+    .Attr("process_set_id: int = 0")
     .Input("tensor: T")
     .Output("output: T")
     .SetShapeFn([](shape_inference::InferenceContext* c) {
@@ -731,6 +842,260 @@ Output
                `tensor` on root rank.
 )doc");
 
+#if TENSORFLOW_VERSION >= 2006000000
+namespace {
+std::string NormalizeNameForTensorFlow(const std::string& name) {
+  static const std::regex normalize_re(R"regex([^a-zA-Z0-9_])regex");
+  return std::regex_replace(name, normalize_re, "_");
+}
+
+Status GetInputDataTypeFromVariable(OpKernelContext* ctx, int input,
+                                    DataType& out) {
+  if (ctx->input_dtype(input) == DT_RESOURCE) {
+    core::RefCountPtr<Var> var;
+    TF_RETURN_IF_ERROR(LookupResource(ctx, HandleFromInput(ctx, input), &var));
+    out = var->tensor()->dtype();
+  } else {
+    out = BaseType(ctx->input_dtype(input));
+  }
+  return Status::OK();
+}
+
+}
+
+template <typename Device>
+class HorovodBroadcastInplaceOp : public OpKernel {
+public:
+  explicit HorovodBroadcastInplaceOp(OpKernelConstruction* context)
+      : OpKernel(context) {
+    OP_REQUIRES_OK(context, context->GetAttr("root_rank", &root_rank_));
+    OP_REQUIRES_OK(context,
+                   context->GetAttr("process_set_id", &process_set_id_));
+    OP_REQUIRES_OK(context, context->GetAttr("num_variables", &num_variables_));
+    OP_REQUIRES_OK(context, context->GetAttr("variable_names", &variable_names_));
+    OP_REQUIRES(context, (int) variable_names_.size() == num_variables_,
+                errors::InvalidArgument(
+                    "len(variable_names) needs to be equal to num_variables"));
+  }
+
+  void Compute(OpKernelContext* context) override {
+    OP_REQUIRES_OK(context, ConvertStatus(common::CheckInitialized()));
+
+    auto any_failures_and_tensors_done =
+        std::make_shared<std::pair<std::atomic<bool>, std::atomic<int>>>();
+    any_failures_and_tensors_done->first.store(false);
+    any_failures_and_tensors_done->second.store(0);
+
+    std::vector<VariableInputLockHolder> variable_locks;
+    variable_locks.reserve(num_variables_);
+
+    for (int tensor_index = 0; tensor_index < num_variables_; ++tensor_index) {
+      DataType dtype;
+      OP_REQUIRES_OK(
+          context, GetInputDataTypeFromVariable(context, tensor_index, dtype));
+
+      // Functions in tensorflow/core/kernels/training_op_helpers.h that deal
+      // with resource variables need a template type parameter. This requires
+      // us to branch out to different specializations of a templated helper
+      // function.
+      switch (dtype) {
+#define PROCESS_CASE(DT, T)                                                    \
+  case DT:                                                                     \
+    OP_REQUIRES_OK(context, Process<T>(context, tensor_index, variable_locks,  \
+                                       any_failures_and_tensors_done));        \
+    break;
+        PROCESS_CASE(DT_UINT8, uint8)
+        PROCESS_CASE(DT_INT8, int8)
+        PROCESS_CASE(DT_INT32, int32)
+        PROCESS_CASE(DT_INT64, int64)
+        PROCESS_CASE(DT_HALF, Eigen::half)
+        PROCESS_CASE(DT_FLOAT, float)
+        PROCESS_CASE(DT_DOUBLE, double)
+        PROCESS_CASE(DT_BOOL, bool)
+        // no support for int16 and uint16 because there are no DenseUpdate
+        // kernels for them
+      default:
+        context->CtxFailure(__FILE__, __LINE__,errors::InvalidArgument(
+            "Horovod inplace broadcast does not support data type ",
+            DataTypeString(dtype)));
+        return;
+      }
+#undef PROCESS_CASE
+    }
+
+    while (!any_failures_and_tensors_done->first.load() &&
+           any_failures_and_tensors_done->second.load() < num_variables_) {
+      std::this_thread::yield();
+    }
+  }
+
+private:
+  int root_rank_ = 0;
+  int process_set_id_ = 0;
+  int num_variables_ = 0;
+  std::vector<std::string> variable_names_;
+
+  template <typename T>
+  Status
+  Process(OpKernelContext* context, int tensor_index,
+          std::vector<VariableInputLockHolder>& variable_locks,
+          const std::shared_ptr<std::pair<std::atomic<bool>, std::atomic<int>>>&
+              any_failures_and_tensors_done) {
+    const bool do_lock = true;
+    const bool sparse = false;
+    // Here we need to replicate the functionality provided by
+    // MaybeLockVariableInputMutexesInOrder(). That function currently does
+    // not work as intended for input_ids not starting at 0. See:
+    // https://github.com/tensorflow/tensorflow/issues/51686
+    {
+      Var* var;
+      mutex* mu = GetTrainingVariableMutex<Device, T>(context, tensor_index,
+                                                      sparse, &var);
+      std::vector<Var*> vars;
+      if (var) {
+        vars.reserve(1);
+        vars.push_back(var);
+      }
+      std::vector<mutex*> mutexes{mu};
+      auto locks = absl::make_unique<std::vector<mutex_lock>>();
+      locks->reserve(1);
+      locks->emplace_back(*mu);
+      auto shared_locks = absl::make_unique<std::vector<tf_shared_lock>>();
+      variable_locks.emplace_back(std::move(vars), std::move(locks),
+                                  std::move(shared_locks));
+    }
+
+    Tensor tensor;
+    TF_RETURN_IF_ERROR(GetInputTensorFromVariable<Device, T>(
+        context, tensor_index, do_lock, sparse, &tensor));
+    Tensor* output = &tensor;
+    MaybeForwardRefInputToRefOutput(context, tensor_index, tensor_index);
+
+    std::string var_name = variable_names_[tensor_index];
+    if (context->input_dtype(tensor_index) == DT_RESOURCE && var_name.empty()) {
+      const ResourceHandle& handle = HandleFromInput(context, tensor_index);
+      // We use handle.name() as a fallback only when we do not have a proper
+      // name because typically it seems to be something like _AnonymousVar18.
+      // The Python name attribute of the variable does not appear to be passed
+      // through automatically.
+      var_name = handle.name();
+    }
+
+    auto device = GetDeviceID(context);
+    // ReadyEvent makes sure input tensor is ready, and output is allocated.
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(
+        std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
+    auto hvd_context = std::make_shared<TFOpContext>(context);
+    auto hvd_tensor = std::make_shared<TFTensor>(tensor);
+    auto hvd_output = std::make_shared<TFTensor>(*output);
+    const std::string node_name =
+        name() + "_" + NormalizeNameForTensorFlow(var_name);
+    auto enqueue_result = EnqueueTensorBroadcast(
+        hvd_context, hvd_tensor, hvd_output, root_rank_, ready_event_list,
+        node_name, device,
+        [context, any_failures_and_tensors_done](const common::Status& status) {
+#if HAVE_GPU
+          auto hvd_event = status.event;
+          if (hvd_event.event) {
+            auto device_context = context->op_device_context();
+            if (device_context != nullptr) {
+              auto stream = stream_executor::gpu::AsGpuStreamValue(
+                  device_context->stream());
+              HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+            }
+          }
+#endif
+          if (!status.ok()) {
+            auto prev_failures = any_failures_and_tensors_done->first.load();
+            if (!prev_failures) {
+              // Only keeping failure status of the first broadcast that fails
+              context->SetStatus(ConvertStatus(status));
+              any_failures_and_tensors_done->first.store(false);
+            }
+          }
+          any_failures_and_tensors_done->second.fetch_add(1);
+        },
+        process_set_id_);
+    return ConvertStatus(enqueue_result);
+  }
+};
+
+REGISTER_KERNEL_BUILDER(Name("HorovodBroadcastInplace").Device(DEVICE_CPU),
+                        HorovodBroadcastInplaceOp<CPUDevice>);
+#if HOROVOD_GPU_BROADCAST
+REGISTER_KERNEL_BUILDER(Name("HorovodBroadcastInplace").Device(DEVICE_GPU),
+                        HorovodBroadcastInplaceOp<GPUDevice>);
+#endif
+
+REGISTER_OP("HorovodBroadcastInplace")
+    .Attr(
+        "T: {uint8, int8, int32, int64, float16, float32, float64, bool}")
+    .Attr("root_rank: int")
+    .Attr("process_set_id: int = 0")
+    .Attr("num_variables: int")
+    .Attr("variable_names: list(string)")
+    .Input("tensor_refs: Ref(num_variables * T)")
+    .Output("output_refs: Ref(num_variables * T)")
+    .SetShapeFn(shape_inference::UnchangedShape)
+    .Doc(R"doc(
+Perform an in-place Broadcast on (TF1-style) reference variables. All other
+processes that do a broadcast on variables with the same names must have the
+same dimensions for those variables. All variables must be located on the same
+device and they must be of the same data type.
+
+This requires TensorFlow 2.6+.
+
+Arguments
+    root_rank:      Rank that will send data, other ranks will receive data.
+    variable_names: Names associated to the variables (obtained via Python
+                    framework)
+
+Input
+    tensor_refs:    Variables to broadcast. They will be updated in-place
+                    to the values from the root rank.
+Output
+    output_refs:    The updated variables.
+)doc");
+
+REGISTER_KERNEL_BUILDER(
+    Name("HorovodBroadcastInplaceResource").Device(DEVICE_CPU),
+    HorovodBroadcastInplaceOp<CPUDevice>);
+#if HOROVOD_GPU_BROADCAST
+REGISTER_KERNEL_BUILDER(Name("HorovodBroadcastInplaceResource")
+                            .Device(DEVICE_GPU)
+                            .HostMemory("resources"),
+                        HorovodBroadcastInplaceOp<GPUDevice>);
+#endif
+
+REGISTER_OP("HorovodBroadcastInplaceResource")
+    .Attr("root_rank: int")
+    .Attr("process_set_id: int = 0")
+    .Attr("num_variables: int")
+    .Attr("variable_names: list(string)")
+    .Input("resources: num_variables * resource")
+    .SetShapeFn(shape_inference::NoOutputs)
+    .Doc(R"doc(
+Perform an in-place Broadcast on (TF2-style) resource variables. All other
+processes that do a broadcast on variables with the same names must have the
+same dimensions for those variables. All variables must be located on the same
+device.
+
+This requires TensorFlow 2.6+.
+
+Arguments
+    root_rank:      Rank that will send data, other ranks will receive data.
+    variable_names: Names associated to the variables (obtained via Python
+                    framework)
+
+Input
+    resources:    Variables to broadcast. They will be updated in-place
+                  to the values from the root rank.
+)doc");
+#endif // TENSORFLOW_VERSION >= 2006000000
+
 class HorovodJoinOp : public AsyncOpKernel {
 public:
   explicit HorovodJoinOp(OpKernelConstruction* context)
@@ -740,12 +1105,30 @@ public:
     OP_REQUIRES_OK_ASYNC(context, ConvertStatus(common::CheckInitialized()),
                          done);
     auto device = GetDeviceID(context);
-    auto ready_event = std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context));
+    Tensor* output = nullptr;
+    OP_REQUIRES_OK_ASYNC(
+        context, context->allocate_output(0, TensorShape(), &output), done);
+
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
     auto hvd_context = std::make_shared<TFOpContext>(context);
+    std::shared_ptr<TFTensor> hvd_output = std::make_shared<TFTensor>(*output);
     auto enqueue_result = EnqueueJoin(
-      hvd_context, ready_event,
+      hvd_context, hvd_output, ready_event_list,
       JOIN_TENSOR_NAME, device,
         [context, done](const common::Status& status) {
+#if HAVE_GPU
+          auto hvd_event = status.event;
+          if (hvd_event.event) {
+            auto device_context = context->op_device_context();
+            if (device_context != nullptr) {
+                auto stream = stream_executor::gpu::AsGpuStreamValue(device_context->stream());
+                HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+            }
+          }
+#endif
           context->SetStatus(ConvertStatus(status));
           done();
         });
@@ -754,18 +1137,106 @@ public:
   }
 };
 
-REGISTER_KERNEL_BUILDER(Name("HorovodJoin").Device(DEVICE_CPU),
-                        HorovodJoinOp);
 #if HOROVOD_GPU_ALLREDUCE
-REGISTER_KERNEL_BUILDER(Name("HorovodJoin").Device(DEVICE_GPU),
+REGISTER_KERNEL_BUILDER(Name("HorovodJoin")
+                            .Device(DEVICE_GPU)
+                            .HostMemory("output"),
+                        HorovodJoinOp);
+#else
+REGISTER_KERNEL_BUILDER(Name("HorovodJoin")
+                            .Device(DEVICE_CPU)
+                            .HostMemory("output"),
                         HorovodJoinOp);
 #endif
 
 REGISTER_OP("HorovodJoin")
+    .Output("output: int32")
+    .SetShapeFn([](shape_inference::InferenceContext* c) {
+      c->set_output(0, c->Scalar());
+      return Status::OK();
+    })
     .Doc(R"doc(
-Perform an join on a tensor,
+Perform a join on a tensor.
+
+Output
+    output:    A scalar integer tensor containing the last rank that joined.
 )doc");
 
+template <typename T, T f(int)>
+class HorovodReturnScalarForProcessSetOp : public OpKernel {
+public:
+  explicit HorovodReturnScalarForProcessSetOp(OpKernelConstruction* context)
+      : OpKernel(context) {
+    OP_REQUIRES_OK(context,
+                   context->GetAttr("process_set_id", &process_set_id_));
+  }
+
+  void Compute(OpKernelContext* context) override {
+    OP_REQUIRES_OK(context, ConvertStatus(common::CheckInitialized()));
+
+    // Write integer to output tensor
+    Tensor* output;
+    OP_REQUIRES_OK(context,
+                   context->allocate_output(0, TensorShape({}), &output));
+
+    auto flat = output->flat<T>();
+    flat(0) = f(process_set_id_);
+  }
+
+private:
+  int process_set_id_;
+};
+
+REGISTER_KERNEL_BUILDER(
+    Name("HorovodSize").Device(DEVICE_CPU).HostMemory("size"),
+    HorovodReturnScalarForProcessSetOp<int, common::horovod_process_set_size>);
+#if HAVE_GPU
+REGISTER_KERNEL_BUILDER(
+    Name("HorovodSize").Device(DEVICE_GPU).HostMemory("size"),
+    HorovodReturnScalarForProcessSetOp<int, common::horovod_process_set_size>);
+#endif
+
+REGISTER_OP("HorovodSize")
+    .Attr("process_set_id: int = 0")
+    .Output("size: int32")
+    .SetIsStateful()
+    .SetShapeFn([](shape_inference::InferenceContext* c) {
+      c->set_output(0, c->Scalar());
+      return Status::OK();
+    })
+    .Doc(R"doc(
+Returns the number of Horovod processes. If process_set_id > 0, limit the
+count to that process set.
+
+Output
+    size:    An integer scalar containing the number of Horovod processes.
+)doc");
+
+REGISTER_KERNEL_BUILDER(
+    Name("HorovodProcessSetIncluded").Device(DEVICE_CPU).HostMemory("included"),
+    HorovodReturnScalarForProcessSetOp<int, common::horovod_process_set_included>);
+#if HAVE_GPU
+REGISTER_KERNEL_BUILDER(
+    Name("HorovodProcessSetIncluded").Device(DEVICE_GPU).HostMemory("included"),
+    HorovodReturnScalarForProcessSetOp<int, common::horovod_process_set_included>);
+#endif
+
+REGISTER_OP("HorovodProcessSetIncluded")
+    .Attr("process_set_id: int = 0")
+    .Output("included: int32")
+    .SetIsStateful()
+    .SetShapeFn([](shape_inference::InferenceContext* c) {
+      c->set_output(0, c->Scalar());
+      return Status::OK();
+    })
+    .Doc(R"doc(
+Returns 0 or 1 depending on whether the current process is
+included in the specified process set or an error code:
+HOROVOD_PROCESS_SET_ERROR_INIT if Horovod is not initialized,
+HOROVOD_PROCESS_SET_ERROR_UNKNOWN_SET if the process set is unknown.
+)doc");
+
+
 template <typename T, T f()> class HorovodReturnScalarOp : public OpKernel {
 public:
   explicit HorovodReturnScalarOp(OpKernelConstruction* context)
@@ -784,29 +1255,6 @@ public:
   }
 };
 
-REGISTER_KERNEL_BUILDER(
-    Name("HorovodSize").Device(DEVICE_CPU).HostMemory("size"),
-    HorovodReturnScalarOp<int, common::horovod_size>);
-#if HAVE_GPU
-REGISTER_KERNEL_BUILDER(
-    Name("HorovodSize").Device(DEVICE_GPU).HostMemory("size"),
-    HorovodReturnScalarOp<int, common::horovod_size>);
-#endif
-
-REGISTER_OP("HorovodSize")
-    .Output("size: int32")
-    .SetIsStateful()
-    .SetShapeFn([](shape_inference::InferenceContext* c) {
-      c->set_output(0, c->Scalar());
-      return Status::OK();
-    })
-    .Doc(R"doc(
-Returns the number of Horovod processes.
-
-Output
-    size:    An integer scalar containing the number of Horovod processes.
-)doc");
-
 REGISTER_KERNEL_BUILDER(
     Name("HorovodLocalSize").Device(DEVICE_CPU).HostMemory("local_size"),
     HorovodReturnScalarOp<int, common::horovod_local_size>);
@@ -886,6 +1334,7 @@ public:
   explicit HorovodAlltoallOp(OpKernelConstruction* context)
       : AsyncOpKernel(context) {
     OP_REQUIRES_OK(context, context->GetAttr("ignore_name_scope", &ignore_name_scope_));
+    OP_REQUIRES_OK(context, context->GetAttr("process_set_id", &process_set_id_));
   }
 
   void ComputeAsync(OpKernelContext* context, DoneCallback done) override {
@@ -902,20 +1351,35 @@ public:
     auto device = GetDeviceID(context);
     auto tensor = context->input(0);
     auto splits = context->input(1);
-    auto ready_event = std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context));
+    common::ReadyEventList ready_event_list;
+#if HAVE_GPU
+    ready_event_list.AddReadyEvent(std::shared_ptr<common::ReadyEvent>(RecordReadyEvent(context)));
+#endif
     auto hvd_context = std::make_shared<TFOpContext>(context);
     auto hvd_tensor = std::make_shared<TFTensor>(tensor);
     auto splits_tensor = std::make_shared<TFTensor>(splits);
     auto enqueue_result = EnqueueTensorAlltoall(
-        hvd_context, hvd_tensor, splits_tensor, ready_event, node_name, device,
+        hvd_context, hvd_tensor, splits_tensor, ready_event_list, node_name, device,
         [context, done](const common::Status& status) {
+#if HAVE_GPU
+          auto hvd_event = status.event;
+          if (hvd_event.event) {
+            auto device_context = context->op_device_context();
+            if (device_context != nullptr) {
+                auto stream = stream_executor::gpu::AsGpuStreamValue(device_context->stream());
+                HVD_GPU_CHECK(gpuStreamWaitEvent(stream, *(hvd_event.event), 0));
+            }
+          }
+#endif
           context->SetStatus(ConvertStatus(status));
           done();
-        });
+        },
+        process_set_id_);
     OP_REQUIRES_OK_ASYNC(context, ConvertStatus(enqueue_result), done);
   }
 private:
   bool ignore_name_scope_;
+  int process_set_id_;
 }; // namespace tensorflow
 
 REGISTER_KERNEL_BUILDER(Name("HorovodAlltoall").Device(DEVICE_CPU),
@@ -932,6 +1396,7 @@ REGISTER_OP("HorovodAlltoall")
     .Attr(
         "T: {uint8, int8, uint16, int16, int32, int64, float16, float32, float64, bool}")
     .Attr("ignore_name_scope: bool = False")
+    .Attr("process_set_id: int = 0")
     .Input("tensor: T")
     .Input("splits: int32")
     .Output("output: T")
